from controlflow.llm.messages import (
    AIMessage,
    AIMessageChunk,
    MessageType,
    ToolCall,
    ToolMessage,
)
from controlflow.utilities.context import ctx
from controlflow.utilities.types import ControlFlowModel


class CompletionEvent(ControlFlowModel):
    type: str
    payload: dict


class CompletionHandler:
    def __init__(self):
        self._response_message_ids: set[str] = set()

    def on_event(self, event: CompletionEvent):
        method = getattr(self, f"on_{event.type}", None)
        if not method:
            raise ValueError(f"Unknown event type: {event.type}")
        method(**event.payload)
        if event.type in [
            "message_done",
            "tool_call_done",
            "tool_result_done",
        ]:
            # only fire the on_response_message hook once per message
            # (a message could contain both a tool call and a message)
            if event.payload["message"].id not in self._response_message_ids:
                self.on_response_message(event.payload["message"])
            self._response_message_ids.add(event.payload["message"].id)

    def on_start(self):
        pass

    def on_end(self):
        pass

    def on_exception(self, exc: Exception):
        pass

    def on_message_created(self, delta: AIMessageChunk):
        pass

    def on_message_delta(self, delta: AIMessageChunk, snapshot: AIMessageChunk):
        pass

    def on_message_done(self, message: AIMessage):
        pass

    def on_tool_call_created(self, delta: AIMessageChunk):
        pass

    def on_tool_call_delta(self, delta: AIMessageChunk, snapshot: AIMessageChunk):
        pass

    def on_tool_call_done(self, message: AIMessage):
        pass

    def on_tool_result_created(self, message: AIMessage, tool_call: ToolCall):
        pass

    def on_tool_result_done(self, message: ToolMessage):
        pass

    def on_response_message(self, message: MessageType):
        """
        This handler is called whenever a message is generated that should be
        included in the completion history (e.g. a `message`, `tool_call` or
        `tool_result`). Note that this is called *in addition* to the respective
        on_*_done handlers, and can be used to quickly collect all messages
        generated during a completion. Messages that satisfy multiple criteria
        (e.g. a message and a tool call) will only be included once.
        """
        pass


class ResponseHandler(CompletionHandler):
    """
    A handler for collecting response messages.
    """

    def __init__(self):
        super().__init__()
        self.response_messages = []

    def on_response_message(self, message: MessageType):
        self.response_messages.append(message)


class TUIHandler(CompletionHandler):
    def on_message_delta(self, delta: AIMessageChunk, snapshot: AIMessageChunk) -> None:
        if tui := ctx.get("tui"):
            tui.update_message(message=snapshot)

    def on_tool_call_delta(self, delta: AIMessageChunk, snapshot: AIMessageChunk):
        if tui := ctx.get("tui"):
            tui.update_message(message=snapshot)

    def on_tool_result_done(self, message: ToolMessage):
        if tui := ctx.get("tui"):
            tui.update_tool_result(message=message)



import asyncio
import contextlib
from typing import TYPE_CHECKING

from prefect.context import FlowRunContext
from prefect.input.run_input import receive_input
from rich.prompt import Prompt as RichPrompt

import controlflow
from controlflow.tools import tool

if TYPE_CHECKING:
    pass


class Prompt(RichPrompt):
    # remove the prompt suffix
    prompt_suffix = " "


async def get_terminal_input(message: str):
    # as a convenience, we wait for human input on the local terminal
    # this is not necessary for the flow to run, but can be useful for testing
    loop = asyncio.get_event_loop()
    user_input = await loop.run_in_executor(
        None,
        RichPrompt.ask,
        f"\n[bold blue]🤖 Agent:[/] [blue]{message}[/]\nType your response",
    )
    return user_input


# async def get_tui_input(tui: "TUIApp", message: str):
#     container = []
#     await tui.get_input(message=message, container=container)
#     while not container:
#         await asyncio.sleep(0.1)
#     return container[0]


async def get_flow_run_input(message: str):
    async for response in receive_input(
        str, flow_run_id=FlowRunContext.get().flow_run.id, poll_interval=0.2
    ):
        return response


@tool
async def talk_to_user(message: str, get_response: bool = True) -> str:
    """
    Send a message to the human user and optionally wait for a response. If
    `get_response` is True, the function will return the user's response,
    otherwise it will return a simple confirmation. Do not send the user
    concurrent messages that require responses, as this will cause confusion.

    You may need to ask the human about multiple tasks at once. Consolidate your
    questions into a single message. For example, if Task 1 requires information
    X and Task 2 needs information Y, send a single message that naturally asks
    for both X and Y.
    """

    if get_response:
        tasks = []
        # if running in a Prefect flow, listen for a remote input
        if (frc := FlowRunContext.get()) and frc.flow_run and frc.flow_run.id:
            remote_input = asyncio.create_task(get_flow_run_input(message=message))
            tasks.append(remote_input)
        # if terminal input is enabled, listen for local input
        if controlflow.settings.enable_local_input:
            local_input = asyncio.create_task(get_terminal_input(message=message))
            tasks.append(local_input)
        if not tasks:
            raise ValueError(
                "No input sources enabled. Either run this task in a flow or enable local input in settings."
            )

        # wait for either the terminal input or the API response
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

        # Get the result of the first completed task
        result = done.pop().result()

        # Cancel any pending tasks
        for task in pending:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task

        return f"User response: {result}"

    return "Message sent to user."
